% FBMPRS_FXN   Function for executing the complex-valued Fast Bayesian 
% Matching Pursuit algorithm using a repeated greedy search (RGS) with
% Q-ary coefficient activity priors that vary across the coefficients
%
% This function allows one to implement the algorithm described in "Fast 
% Bayesian Matching Pursuit: Model Uncertainty and Parameter Estimation 
% for Sparse Linear Models," P. Schniter, L. C. Potter, and J. Ziniel.
% FBMPCS_FXN will find a high-probability set of sparse solutions "x"
% to the underdetermined system of linear equations
%                           y = Ax + w,
% where both x and w are real-valued Gaussian. 
%
% SYNTAX: [xmmse, llr, xmmse_star, psy_star, nu_star, T_star, d_tot, 
%	d_max, Cov_star] = fbmprs_fxn(y, A, lam1, sig2w, sig2q, muq, 
%					llrmax, D, stop, psy_min)
%
% INPUTS:
%
% y -       Vector of observations of length M
%
% A -       M-by-N measurement matrix 
%
% lam1 -    (Q-1)-by-N matrix of activity priors (i.e. Pr{s_n = q} = lam1(q,n),
%           q > 0)
%
% sig2w -   Noise variance
%
% sig2q -   2-by-1 vector of sparse coefficient variances,  ([off; on]).  
%           Note that the variances of all Q-1 active Gaussian mixture
%           densities will be equal
%
% muq -     Q-by-1 vector of means for the Q-ary Gaussian mixture 
%           densities, ([off; on_1; on_2; ...; on_Q-1]).  Note that the 
%           first element of the vector should be zero, to model sparsity.
%
% llrmax -  Maximum absolute value for reported posterior LLRs (default 5)
%
% D -       Maximum number of repeated greedy searches (default 5)
%
% stop -    Threshold for the number of standard deviations below
%           E[nu(s,y)] at which FBMP may terminate upon finding a metric
%           that exceeds it, (default 0)
%
% psy_min - Value of smallest reported posterior relative to largest
%	    (defaul 1e-4)
%
% OUTPUTS:
%
% xmmse -       Approximate sparse MMSE estimate of x, given y
%
% llr -		Vector of posterior log-likelihood ratios
%
% xmmse_star -	Significant conditional MMSE estimates of x, given s and y
%
% psy_star -    Corresponding significant conditional mixture vector
%               posterior probabilities, p(s|y)
%
% nu_star -     Corresponding metric values, nu(s,y)
%
% T_star -      Indices of active taps for the conditional MMSE estimates 
%               of x returned in xmmse_star
%
% d_tot -       Number of required RGS iterations to find a mixture vector
%               that exceeded E[nu(s,y)] - stop*stddev(nu(s,y)).  If no 
%               such vector is found, 'inf' is returned
%
% d_max -       RGS iteration at which the maximum metric was found, 
%               (1 <= d_max <= d_tot)
%
% Cov_star -    Significant conditional covariance matrices, given s and y 
%

%
% Coded by: Philip Schniter, The Ohio State Univ.
% Modified by: Justin Ziniel, The Ohio State Univ.
% E-mail: schniter@ece.osu.edu, ziniel.1@osu.edu
% Last change: April 11, 2012
% Change summary: Created from fbmprs_fxn.m (04/11/12)
% FBMP version 1.3 
% Copyright (c) Philip Schniter, Lee C. Potter, and Justin Ziniel, 2008
%

function [xmmse, llr, xmmse_star, psy_star, nu_star, T_star, d_tot, d_max, ...
         Cov_star] = fbmprs_fxn(y, A, lam1, sig2w, sig2q, muq, llrmax, D, ...
	 			stop, psy_min)


%% Fill in missing variables and check input for errors

% Set defaults
if (nargin < 10)
    psy_min = 1e-4;		% minimum posterior relative to largest one found 
    if (nargin < 9)
        stop = 0;		% Stop at zero std. devs. below the mean of nu(s,y)
        if (nargin < 8)
            D = 5;		% Maximum number of RGS iterations
    	    if (nargin < 7)
	        llrmax = 5;	% Maximum absoluate value of reported LLRs
                if (nargin < 6)
                    error('FBMPRS_FXN requires at least 6 arguments')
                end
	    end
        end
    end
end;

if (nargout < 9)
    calc_cov = 0;		% do not calculate covariance matrices
else
    calc_cov = 1;
end;


[M, N] = size(A);
sig2q = sig2q(:);
muq = muq(:);
if (size(lam1,2)~=N) && (size(lam1,1)==N), lam1 = lam1'; end;

if (length(y) ~= M)
    error('y and A have a differing number of rows')
elseif (length(sig2q) ~= 2)
    error('sig2q must be a vector of length 2')
elseif ( (size(lam1,2)~=N) || (size(lam1,1)~=length(muq)-1) )
    error('Sizes of A, lam1, and muq must match')
elseif ( sum( sum(lam1,1) < 0 ) || sum( sum(lam1,1) > 1 ) )
    error('lam1 must contain admissible probability values')
elseif ( (min(sig2q) < 0) || (sig2w < 0) )
    error('Elements of sig2q and sig2w must be non-negative')
elseif (D < 1)
    error('D must be >= 1')
elseif (min(sig2q(2:end)) < 1e-6)
    error('Elements in sig2q(2:end) must be at least 1e-6')
end

Q = length(muq) - 1;        % # of Gaussian mixture densities minus one
sig2q = [sig2q; sig2q(2)*ones(Q-1,1)];  % Equal variance for all active Gaussians
lam = [1-sum(lam1,1);lam1];
K = sum(sum(lam1));

% 1st and 2nd moments of mixture selection metric prior distribution
a2 = trace(A'*A)/N;		% average 2-norm of the columns of A
loglam = sum(lam.*log(lam+eps),1);
nu_true_mean = -M/2 -M/2*log(sig2w*2*pi) -K/2*log(a2*sig2q(2)/sig2w+1) +sum(loglam);
nu_true_stdv = sqrt( M/2 + sum( sum(lam.*(log(lam+eps)-ones(Q+1,1)*loglam).^2,1) ...
	+(1/2*log(a2*sig2q(2)/sig2w+1))^2*lam(1,:).*(1-lam(1,:)) ...
	-log(a2*sig2q(2)/sig2w+1)*lam(1,:).*(loglam-log(lam(1,:)+eps)) ) );
nu_stop = nu_true_mean - stop*nu_true_stdv;


% Default algorithmic parameters
P = min(M,1+ceil(K+erfcinv(1e-2)*sqrt(2*lam(1,:)*(1-lam(1,:))')));% search length


%% Repeated greedy search

% Allocate variables for storage
T = cell(P,D);          % indices of active taps
sT = cell(P,D);         % active mixing params
nu = -inf*ones(P,D);	% metrics
xmmse = cell(P,D);		% mmse conditioned on active taps
Cov = cell(P,D);		% mmse-covariance conditioned on active taps
d_tot = inf;            % flag maximum number of RGS iterations at infinity

% Initialize (root node)
nu_root = -norm(y)^2/2/sig2w - M*log(2*pi)/2 - M*log(sig2w)/2 + sum(log(lam(1,:)+eps));
Bxt_root = A/sig2w;
betaxt_root = abs(sig2q(2)*(1 + sig2q(2)*sum(conj(A).*Bxt_root)).^(-1));
nuxt_root = zeros(1,Q*N);	% Q = # non-zero means
for q = 1:Q
    nuxt_root([1:N]+(q-1)*N) = nu_root + log(betaxt_root/sig2q(2))/2 ...
            + 1/2*betaxt_root.*abs( y'*Bxt_root + muq(q + 1)/sig2q(2)).^2 ...
            - 1/2*abs(muq(q + 1))^2/sig2q(2) + log(lam(q+1,:)+eps) - log(lam(1,:)+eps);
end

% Descend one branch at a time
for d = 1:D,
    nuxt = nuxt_root;
    z = y;
    Bxt = Bxt_root;
    betaxt = betaxt_root;
    for p = 1:P
        [nustar,nqstar] = max(nuxt);                % find best extension
        while sum(abs(nustar-nu(p,1:d-1)) < 1e-8)   % if same as explored node...
            nuxt(nqstar) = -inf;                    % ... mark extension as redundant
            [nustar, nqstar] = max(nuxt);           % ... and find next best extension
        end
        qstar = floor((nqstar - 1)/N) + 1;          % mean index of best extension
        nstar = mod(nqstar - 1, N) + 1;             % coef index of best extension
        nu(p,d) = nustar;                           % replace worst explored node...
        if (p > 1)
            T{p,d} = [T{p-1,d}, nstar];
            sT{p,d} = [sT{p-1,d}, 1 + qstar];
        else
            T{p,d} = nstar;
            sT{p,d} = 1 + qstar;
        end
        z = z - A(:,nstar)*muq(qstar+1);
        Bxt = Bxt - Bxt(:,nstar)*betaxt(nstar)*( Bxt(:,nstar)'*A );
        xmmse{p,d} = zeros(N,1);
        xmmse{p,d}(T{p,d}) = muq(sT{p,d}) + sig2q(2)*Bxt(:,T{p,d})'*z;
	if calc_cov,
          Cov{p,d} = sig2q(2)*eye(length(T{p,d})) - ...
                        sig2q(2)^2*Bxt(:,T{p,d})'*A(:,T{p,d});
	end;
        betaxt = abs(sig2q(2)*(1 + sig2q(2)*sum(conj(A).*Bxt)).^(-1));
        for q = 1:Q                                 % Q = # non-zero means
            nuxt([1:N]+(q-1)*N) = nu(p,d) + log(betaxt/sig2q(2))/2 ...
                    + 1/2*betaxt.*abs( z'*Bxt+muq(q+1)/sig2q(2) ).^2 ...
                    - 1/2*abs(muq(q+1))^2/sig2q(2)  ...
		    + log(lam(q+1,:)+eps) - log(lam(1,:)+eps);
            % can't activate an already activated coefficient!
            nuxt(T{p,d}+(q-1)*N) = -inf*ones(size(T{p,d}));
        end
    end

    if (max(nu(:,d)) > nu_stop)     % A mixture vector has exceeded the threshold
        d_tot = d;
        break
    end
end
nu = nu(:,1:d);


%% Calculate stuff for statistically significant hypotheses

[dum,indx] = sort(nu(:), 'descend');
d_max = ceil(indx(1)/P);
nu_max = nu(indx(1));
num = sum( nu(:) > nu_max+log(psy_min) );
nu_star = nu(indx(1:num));
T_star = cell(1,num);
xmmse_star = cell(1,num);           % mmse est cond on mixing params
Cov_star = cell(1,num);             % Cov cond on mixing params
for k = 1:num
    T_star{k} = T{indx(k)};
    xmmse_star{k} = xmmse{indx(k)};
    Cov_star{k} = Cov{indx(k)};
end
psy_star = exp(nu_star - nu_max)/sum(exp(nu_star - nu_max));    % posteriors
xmmse = [xmmse_star{:}]*psy_star;  % approximate mmse estimate

%% Calculate posterior LLRs
s_star = ([xmmse_star{:}]~=0);		% active coefs
s_star_not = not(s_star);		% inactive coefs
llr = log((s_star*exp(nu_star - nu_max))./(s_star_not*exp(nu_star - nu_max)));
llr = max(min(llr,llrmax),-llrmax);	% quantize

